# pylint: disable=E0611,C0103,R0914,W1514,R1735,R1734,C0206

import asyncio
from typing import Dict, List
from time import time
from zipfile import ZipFile
import tempfile
import json
import os
import shutil

import numpy as np
from ..pytpu import TPUDevice, TPUProgram, TPUInference, TPUProgramInfo  # type: ignore
from ..tools.helpers import to_raw

# tpu_program_path = '/auto/pytpu_tests/tpu_programs/tpu_framework_3.0.0'
# tpu_program = 'resnet50.tpu'

__all__ = [
    'get_fps',
]


async def run_inference(name: int, device: TPUDevice, inference: TPUInference,
                        data_list: List[Dict[str, np.ndarray]], silent: bool = True) -> None:

    for ii, data in enumerate(data_list):
        inference.load(data)
        if not silent:
            print(f'Process {name}: Load tensors {ii} done')
        status = await device.load_inference(inference)
        assert status.is_success
        if not silent:
            print(f'Process {name}: Run {ii} done')
        inference.get(as_dict=True)
        if not silent:
            print(f'Process {name}: Get tensors {ii} done')

        if (ii + 1) % 100 == 0:
            print(f'Process {name}: finish {ii + 1} iteration')


def get_fps(program_path: str, raw: bool = False, n_queries: int = 100, n_proc: int = 4) -> float:

    print(f'Start measure performance for program: {program_path}')
    print(f'Configuration: RAW = {raw}; queries = {n_queries}; processes = {n_proc}')

    with tempfile.TemporaryDirectory() as tempdir:
        with ZipFile(program_path, 'r') as zip_obj:
            zip_obj.extractall(tempdir)

        with open(os.path.join(tempdir, 'metadata.json'), 'r') as metadata_file:
            metadata = json.load(metadata_file)

        if raw is True:
            with open(os.path.join(tempdir, 'metadata.json'), 'w') as metadata_file:
                metadata = to_raw(metadata)
                json.dump(metadata, metadata_file)

            program_path = os.path.join(tempdir, 'program_raw.tpu')
            shutil.make_archive(program_path, 'zip', tempdir)
            os.rename(program_path + '.zip', program_path)
            print(f'Raw program saved to {program_path}')

        layers_param = dict()
        for _, region in metadata['inputs'].items():
            for name, inp in region.items():
                layers_param[inp['anchor']] = inp

        data_list = list()
        for _ in range(n_queries):
            for name in layers_param:
                # print(layers_param[name]['user_shape'])
                generated_data = (np.random.rand(*layers_param[name]['user_shape']) * 2 - 1) * 100
                # print(layers_param[name]['user_dtype'])
                generated_data = generated_data.astype(layers_param[name]['user_dtype'])
                generated_data_dict = {name: generated_data}
                data_list.append(generated_data_dict)

        batch = max([layers_param[name]['user_shape'][0] for name in layers_param])

        device = TPUDevice()
        tpu_program = TPUProgram(program_path, TPUProgramInfo())
        device.load_program(tpu_program)
        inferences = [TPUInference(tpu_program) for _ in range(n_proc)]
        inference_processes = [run_inference(ii, device, inferences[ii], data_list[ii:len(data_list):n_proc])
                               for ii in range(n_proc)]

        start_time = time()
        asyncio.get_event_loop().run_until_complete(asyncio.gather(*inference_processes))
        total_inference_time = time() - start_time

        fps = n_queries * batch / total_inference_time

        print(f'Estimated FPS = {fps}')

    return fps



import argparse
from pytpu.tools import get_fps


def main():
    parser = argparse.ArgumentParser(prog='run_get_fps',
                                     description='TPU performance measurement tool')

    parser.add_argument('program', nargs='?', help='Path to the program.tpu', type=str)
    parser.add_argument('-raw', nargs='?', help='Enable RAW mode without pre- and post-processing (default = False)',
                        type=int, default=False)
    parser.add_argument('-n_queries', nargs='?', help='Number of queries (default = 1000)', type=int, default=100)
    parser.add_argument('-n_proc', nargs='?', help='Number of processes (default = 4)', type=int, default=4)

    args = parser.parse_args()

    get_fps(args.program, raw=bool(args.raw), n_queries=args.n_queries, n_proc=args.n_proc)



# import importlib.util
# import sys
from .tools import get_fps
# pylint: disable=E0611
from .pytpu import ProcessingMode  # type: ignore
from .pytpu import TPUDevice  # type: ignore
from .pytpu import TPUInference  # type: ignore
from .pytpu import TPUInferenceException  # type: ignore
from .pytpu import TPUProgram  # type: ignore
from .pytpu import TPUProgramException  # type: ignore
from .pytpu import TPUProgramInfo  # type: ignore
# pylint: enable=E0611

__all__ = [
    'get_fps',
    'ProcessingMode',
    'TPUDevice',
    'TPUInference',
    'TPUInferenceException',
    'TPUProgram',
    'TPUProgramException',
    'TPUProgramInfo',
]

# spec = importlib.util.spec_from_file_location("libtpu", "/usr/lib/libtpu.so")
# libtpu = importlib.util.module_from_spec(spec)
# sys.modules['pytpu'] = libtpu
# sys.modules["pytpu"] = sys.modules["pytpu"].append(libtpu)



# coding: utf-8
"""Setup script for IVA TPU."""
from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

# tmp = find_packages()

setup(name='pytpu',
      packages=['pytpu', 'pytpu.tools', 'pytpu.pytpu', 'pytpu.scripts'],
      # packages=find_packages(),
      version="12.0.0",
      author="Maxim Moroz",
      author_email="m.moroz@iva-tech.ru",
      description="TPU Python API",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="http://git.mmp.iva-tech.ru/tpu_sw/iva_tpu_sdk",
      install_requires=[
            'numpy>=1.14',
      ],
      zip_safe=False,
      python_requires='>=3.6',
      entry_points={
            'console_scripts': [
                'run_get_fps = pytpu.scripts.run_get_fps:main'
            ]
        },
      )



import copy
from typing import Dict


def to_raw(metadata: Dict) -> Dict:
    raw_metadata = copy.copy(metadata)
    for idx, region in raw_metadata['inputs'].items():
        for name, _ in region.items():
            raw_metadata['inputs'][idx][name]['user_shape'] = raw_metadata['inputs'][idx][name]['tpu_shape']
            raw_metadata['inputs'][idx][name]['user_order'] = raw_metadata['inputs'][idx][name]['tpu_order']
            raw_metadata['inputs'][idx][name]['user_dtype'] = raw_metadata['inputs'][idx][name]['tpu_dtype']
            raw_metadata['inputs'][idx][name]['scales'] = [1.0, ]
            raw_metadata['inputs'][idx][name]['padding'] = \
                [[0, 0], ] * len(raw_metadata['inputs'][idx][name]['tpu_shape'])

    for idx, region in raw_metadata['outputs'].items():
        for name, _ in region.items():
            raw_metadata['outputs'][idx][name]['user_shape'] = raw_metadata['outputs'][idx][name]['tpu_shape']
            raw_metadata['outputs'][idx][name]['user_order'] = raw_metadata['outputs'][idx][name]['tpu_order']
            raw_metadata['outputs'][idx][name]['user_dtype'] = raw_metadata['outputs'][idx][name]['tpu_dtype']
            raw_metadata['outputs'][idx][name]['scales'] = [1.0, ]
            raw_metadata['outputs'][idx][name]['padding'] = \
                [[0, 0], ] * len(raw_metadata['outputs'][idx][name]['tpu_shape'])

    return raw_metadata
